(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
  typeof define === 'function' && define.amd ? define(factory) :
  (global = global || self, global.nestedSort = factory());
}(this, (function () { 'use strict';

  function _classCallCheck(instance, Constructor) {
    if (!(instance instanceof Constructor)) {
      throw new TypeError("Cannot call a class as a function");
    }
  }

  function _defineProperties(target, props) {
    for (var i = 0; i < props.length; i++) {
      var descriptor = props[i];
      descriptor.enumerable = descriptor.enumerable || false;
      descriptor.configurable = true;
      if ("value" in descriptor) descriptor.writable = true;
      Object.defineProperty(target, descriptor.key, descriptor);
    }
  }

  function _createClass(Constructor, protoProps, staticProps) {
    if (protoProps) _defineProperties(Constructor.prototype, protoProps);
    if (staticProps) _defineProperties(Constructor, staticProps);
    return Constructor;
  }

  function _defineProperty(obj, key, value) {
    if (key in obj) {
      Object.defineProperty(obj, key, {
        value: value,
        enumerable: true,
        configurable: true,
        writable: true
      });
    } else {
      obj[key] = value;
    }

    return obj;
  }

  function _toConsumableArray(arr) {
    return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _unsupportedIterableToArray(arr) || _nonIterableSpread();
  }

  function _arrayWithoutHoles(arr) {
    if (Array.isArray(arr)) return _arrayLikeToArray(arr);
  }

  function _iterableToArray(iter) {
    if (typeof Symbol !== "undefined" && Symbol.iterator in Object(iter)) return Array.from(iter);
  }

  function _unsupportedIterableToArray(o, minLen) {
    if (!o) return;
    if (typeof o === "string") return _arrayLikeToArray(o, minLen);
    var n = Object.prototype.toString.call(o).slice(8, -1);
    if (n === "Object" && o.constructor) n = o.constructor.name;
    if (n === "Map" || n === "Set") return Array.from(n);
    if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
  }

  function _arrayLikeToArray(arr, len) {
    if (len == null || len > arr.length) len = arr.length;

    for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];

    return arr2;
  }

  function _nonIterableSpread() {
    throw new TypeError("Invalid attempt to spread non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
  }

  var DataEngine = /*#__PURE__*/function () {
    /**
     * @constructor
     * @param {object[]} data
     * @param {object} [propertyMap={}]
     */
    function DataEngine(_ref) {
      var data = _ref.data,
          _ref$propertyMap = _ref.propertyMap,
          propertyMap = _ref$propertyMap === void 0 ? {} : _ref$propertyMap;

      _classCallCheck(this, DataEngine);

      this.data = data;
      this.sortedData = [];
      this.sortedDataDomArray = [];
      this.propertyMap = propertyMap;
      this.maybeTransformData();
    }

    _createClass(DataEngine, [{
      key: "maybeTransformData",
      value: function maybeTransformData() {
        if (!Object.keys(this.propertyMap).length) return;
        var getItemPropProxyName = this.getItemPropProxyName.bind(this);
        this.data = this.data.map(function (item) {
          return new Proxy(item, {
            get: function get(target, prop, receiver) {
              return Reflect.get(target, getItemPropProxyName(prop), receiver);
            }
          });
        });
      }
      /**
       * @param {PropertyKey} prop
       * @returns {PropertyKey}
       */

    }, {
      key: "getItemPropProxyName",
      value: function getItemPropProxyName(prop) {
        if (this.propertyMap.hasOwnProperty(prop)) {
          return this.propertyMap[prop];
        }

        return prop;
      }
      /**
       * @returns {object[]}
       */

    }, {
      key: "sortListItems",
      value: function sortListItems() {
        this.sortedData = _toConsumableArray(this.data).sort(function (item1, item2) {
          if (!item1.parent && item2.parent) return -1;
          return !item2.parent && item1.parent ? 1 : 0;
        });
        return this.sortedData;
      }
      /**
       * @param {object[]} item
       * @param {string} nodeName
       * @returns {HTMLElement}
       */

    }, {
      key: "createItemElement",
      value: function createItemElement(item) {
        var nodeName = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 'li';
        var id = item.id,
            text = item.text;
        var el = document.createElement(nodeName);
        el.dataset.id = id;
        if (nodeName === 'li') el.innerHTML = text;
        return el;
      }
      /**
       * @param {HTMLElement} node
       * @param {object} item
       * @returns {boolean}
       */

    }, {
      key: "elementIsParentOfItem",
      value: function elementIsParentOfItem(node, item) {
        return node.dataset.id === "".concat(item.parent);
      }
      /**
       * @param {HTMLElement} node
       * @param {object} item
       * @param {string} nodeName
       * @returns {HTMLElement|null}
       */

    }, {
      key: "getParentNodeOfItem",
      value: function getParentNodeOfItem(node, item, nodeName) {
        return node.querySelector("".concat(nodeName, "[data-id=\"").concat(item.parent, "\"]"));
      }
      /**
       * @param {HTMLElement} node
       * @param {object} item
       * @returns {boolean}
       */

    }, {
      key: "elementIsAncestorOfItem",
      value: function elementIsAncestorOfItem(node, item) {
        return !!this.getParentNodeOfItem(node, item, 'li');
      }
      /**
       * @param {HTMLElement} node
       * @param {object} item
       * @returns {HTMLElement}
       */

    }, {
      key: "getDirectListParentOfItem",
      value: function getDirectListParentOfItem(node, item) {
        return this.getParentNodeOfItem(node, item, 'ul');
      }
      /**
       * @param {object} item
       * @returns {boolean}
       */

    }, {
      key: "maybeAppendItemToParentDom",
      value: function maybeAppendItemToParentDom(item) {
        var _this = this;

        var parent = item.parent;
        var topParent = this.sortedDataDomArray.find(function (topLevelListItem) {
          return _this.elementIsParentOfItem(topLevelListItem, item) || _this.elementIsAncestorOfItem(topLevelListItem, item);
        });
        if (!topParent) return false;
        var listItem = this.createItemElement(item);
        var directParentList = this.getDirectListParentOfItem(topParent, item);

        if (!directParentList) {
          // so we need to create the direct parent UL and append it to the direct parent LI
          directParentList = this.createItemElement({
            id: parent
          }, 'ul');
          var directParentListItem = this.getParentNodeOfItem(topParent, item, 'li') || topParent;
          directParentListItem.appendChild(directParentList);
        }

        directParentList.appendChild(listItem);
        return true;
      }
      /**
       * @returns {array}
       */

    }, {
      key: "getListItemsDom",
      value: function getListItemsDom() {
        var _this2 = this;

        this.sortedDataDomArray = [];
        var processedItems = [];

        while (processedItems.length !== this.sortListItems().length) {
          processedItems = this.sortedData.reduce(function (processedItems, item) {
            var id = item.id;
            if (processedItems.includes(id)) return processedItems;
            var itemAdded;

            if (!item.parent) {
              var listItem = _this2.createItemElement(item);

              _this2.sortedDataDomArray.push(listItem);

              itemAdded = true;
            } else {
              itemAdded = _this2.maybeAppendItemToParentDom(item);
            }

            if (itemAdded) processedItems.push(id);
            return processedItems;
          }, processedItems);
        }

        return this.sortedDataDomArray;
      }
      /**
       * @param {HTMLUListElement} ul
       * @returns {object[]}
       */

    }, {
      key: "convertDomToData",
      value: function convertDomToData(ul) {
        var _this3 = this;

        return Array.from(ul.querySelectorAll('li')).map(function (li) {
          var _ref2;

          var parentListItem = li.parentNode;
          var parent = parentListItem.dataset.id;
          return _ref2 = {}, _defineProperty(_ref2, _this3.getItemPropProxyName('id'), li.dataset.id), _defineProperty(_ref2, _this3.getItemPropProxyName('parent'), parent), _ref2;
        });
      }
      /**
       * @returns {HTMLUListElement}
       */

    }, {
      key: "render",
      value: function render() {
        var list = document.createElement('ul');
        this.getListItemsDom().forEach(function (listItem) {
          return list.appendChild(listItem);
        });
        return list;
      }
    }]);

    return DataEngine;
  }();

  var nestedSort = /*#__PURE__*/function () {
    /**
     * @constructor
     * @param {{onDrop: function}} actions
     * @param {array} data
     * @param {number} droppingEdge
     * @param {string} el
     * @param {array|string} listClassNames
     * @param {object} [propertyMap={}]
     */
    function nestedSort() {
      var _ref = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
          _ref$actions = _ref.actions;

      _ref$actions = _ref$actions === void 0 ? {} : _ref$actions;
      var onDrop = _ref$actions.onDrop,
          data = _ref.data,
          _ref$droppingEdge = _ref.droppingEdge,
          droppingEdge = _ref$droppingEdge === void 0 ? 15 : _ref$droppingEdge,
          el = _ref.el,
          listClassNames = _ref.listClassNames,
          _ref$propertyMap = _ref.propertyMap,
          propertyMap = _ref$propertyMap === void 0 ? {} : _ref$propertyMap;

      _classCallCheck(this, nestedSort);

      this.data = data;
      this.selector = el;
      this.sortableList = null;
      this.placeholderUl = null;
      this.placeholderInUse = null;
      this.draggedNode = null;
      this.targetedNode = null;
      this.listClassNames = this.createListClassNamesArray(listClassNames);
      this.propertyMap = propertyMap;
      this.actions = {
        onDrop: onDrop
      };
      this.targetNode = {
        X: null,
        Y: null
      };
      this.distances = {
        droppingEdge: droppingEdge,
        droppingEdgeNegative: droppingEdge * -1,
        mouseTo: {
          targetedElTop: undefined
        }
      };
      this.dimensions = {
        targetedEl: {
          H: undefined
        }
      };
      this.cursor = {
        X: null,
        Y: null
      };
      this.classNames = {
        dragged: 'ns-dragged',
        placeholder: 'ns-placeholder',
        targeted: 'ns-targeted'
      };
      this.maybeInitDataDom();
      this.addListAttributes();
      this.initDragAndDrop();
    }

    _createClass(nestedSort, [{
      key: "getDataEngine",
      value: function getDataEngine() {
        if (this.dataEngine instanceof DataEngine) {
          return this.dataEngine;
        }

        this.dataEngine = new DataEngine({
          data: this.data,
          propertyMap: this.propertyMap
        });
        return this.dataEngine;
      }
    }, {
      key: "createListClassNamesArray",
      value: function createListClassNamesArray(listClassNames) {
        if (!listClassNames) return [];
        return Array.isArray(listClassNames) ? listClassNames : listClassNames.split(' ');
      }
    }, {
      key: "maybeInitDataDom",
      value: function maybeInitDataDom() {
        if (!(Array.isArray(this.data) && this.data.length)) return;
        var list = this.getDataEngine().render();
        document.querySelector(this.selector).appendChild(list);
      }
    }, {
      key: "getSortableList",
      value: function getSortableList() {
        if (this.sortableList instanceof HTMLUListElement) return this.sortableList;

        if (this.selector instanceof HTMLUListElement) {
          this.sortableList = this.selector;
        } else {
          var list = document.querySelector(this.selector);
          this.sortableList = list.nodeName === 'UL' ? list : list.querySelector('ul');
        }

        return this.sortableList;
      }
    }, {
      key: "addListAttributes",
      value: function addListAttributes() {
        var _this$getSortableList;

        (_this$getSortableList = this.getSortableList().classList).add.apply(_this$getSortableList, _toConsumableArray(this.listClassNames));
      }
    }, {
      key: "initDragAndDrop",
      value: function initDragAndDrop() {
        var _this = this;

        document.addEventListener('dragover', this.dragListener.bind(this), false);
        this.initPlaceholderList();
        this.getSortableList().querySelectorAll('li').forEach(function (el) {
          el.setAttribute('draggable', 'true');
          el.addEventListener('dragstart', _this.onDragStart.bind(_this), false);
          el.addEventListener('dragenter', _this.onDragEnter.bind(_this), false);
          el.addEventListener('dragover', _this.onDragOver.bind(_this), false);
          el.addEventListener('dragleave', _this.onDragLeave.bind(_this), false);
          el.addEventListener('dragend', _this.onDragEnd.bind(_this), false);
          el.addEventListener('drop', _this.onDrop.bind(_this), false);

          _this.addListItemStyles(el);
        });
      }
    }, {
      key: "getComputedStyleValue",
      value: function getComputedStyleValue(el, prop) {
        return window.getComputedStyle(el, null).getPropertyValue(prop);
      }
    }, {
      key: "addListItemStyles",
      value: function addListItemStyles(li) {
        // let's add a move cursor icon if it does not already have a cursor css property
        var cursor = this.getComputedStyleValue(li, 'cursor');

        if (!cursor || cursor === 'auto') {
          li.style.cursor = 'move';
        }
      }
    }, {
      key: "onDragStart",
      value: function onDragStart(e) {
        this.draggedNode = e.target;
        this.draggedNode.classList.add(this.classNames.dragged);
        e.dataTransfer.setData('text', 'Drag started!'); // Hack for Firefox!
      }
    }, {
      key: "onDragOver",
      value: function onDragOver(e) {
        e.preventDefault(); // prevent default to allow drop
      }
    }, {
      key: "onDragEnter",
      value: function onDragEnter(e) {
        if (!this.draggedNode) return;

        if (['LI', 'UL'].includes(e.target.nodeName)) {
          e.preventDefault(); // prevent default to allow drop

          if (this.targetedNode) this.targetedNode.classList.remove(this.classNames.targeted);
          this.targetedNode = e.target;
          e.target.classList.add(this.classNames.targeted);
        }
      }
    }, {
      key: "onDragLeave",
      value: function onDragLeave(e) {
        e.preventDefault();
        e.target.removeEventListener('dragover', this.onDrop);
        e.target.removeEventListener('drop', this.onDrop);
        e.target.removeEventListener('dragleave', this.onDragLeave);
      }
    }, {
      key: "onDragEnd",
      value: function onDragEnd(e) {
        e.preventDefault();
        e.stopPropagation();
        this.draggedNode.classList.remove(this.classNames.dragged);
        this.targetedNode.classList.remove(this.classNames.targeted);
        this.cleanupPlaceholderLists();
        this.draggedNode = null;
      }
    }, {
      key: "onDrop",
      value: function onDrop(e) {
        e.preventDefault();
        e.stopPropagation();
        this.maybeDrop();
        this.cleanupPlaceholderLists();

        if (typeof this.actions.onDrop === 'function') {
          this.actions.onDrop(this.getDataEngine().convertDomToData(this.getSortableList()));
        }
      }
    }, {
      key: "dragListener",
      value: function dragListener(e) {
        this.updateCoordination(e);
        this.managePlaceholderLists(e);
      }
    }, {
      key: "updateCoordination",
      value: function updateCoordination(e) {
        this.calcMouseCoords(e);
        this.calcMouseToTargetedElDist();
      }
    }, {
      key: "maybeDrop",
      value: function maybeDrop(e) {
        if (!this.canBeDropped()) {
          return;
        }

        var dropLocation;

        if (this.targetedNode.nodeName === 'LI' && !this.cursorIsIndentedEnough()) {
          dropLocation = 'before';
        } else if (this.targetedNode.nodeName === 'UL') {
          dropLocation = 'inside';
        }

        if (dropLocation) this.dropTheItem(dropLocation, e);
      }
    }, {
      key: "dropTheItem",
      value: function dropTheItem(place, e) {
        switch (place) {
          case 'before':
            this.targetedNode.parentNode.insertBefore(this.draggedNode, this.targetedNode);
            break;

          case 'after':
            this.insertAfter(this.draggedNode, this.targetedNode);
            break;

          case 'inside':
            this.targetedNode.appendChild(this.draggedNode);
            break;
        }
      }
    }, {
      key: "calcMouseCoords",
      value: function calcMouseCoords(e) {
        // we're having the client coords because on the next lines, we use getBoundingClientRect which behaves in the same way
        this.cursor.X = e.clientX;
        this.cursor.Y = e.clientY;
      }
    }, {
      key: "calcMouseToTargetedElDist",
      value: function calcMouseToTargetedElDist() {
        if (!this.targetedNode) {
          return;
        }

        var offset = this.targetedNode.getBoundingClientRect();
        this.targetNode.X = offset.left;
        this.targetNode.Y = offset.top;
        var result = this.targetNode.Y - this.cursor.Y;
        this.distances.mouseTo.targetedElTop = result;
        this.distances.mouseTo.targetedElTopAbs = Math.abs(result);
        this.dimensions.targetedEl.H = this.targetedNode.clientHeight;
        this.distances.mouseTo.targetedElBot = this.distances.mouseTo.targetedElTopAbs - this.dimensions.targetedEl.H;
      }
    }, {
      key: "areNested",
      value: function areNested(child, parent) {
        return parent && Array.from(parent.querySelectorAll('li')).some(function (li) {
          return li === child;
        });
      }
    }, {
      key: "cursorIsIndentedEnough",
      value: function cursorIsIndentedEnough() {
        return this.cursor.X - this.targetNode.X > 50;
      }
    }, {
      key: "mouseIsTooCloseToTop",
      value: function mouseIsTooCloseToTop() {
        return this.cursor.Y - this.targetNode.Y < this.distances.droppingEdge;
      }
    }, {
      key: "insertAfter",
      value: function insertAfter(newNode, referenceNode) {
        referenceNode.parentNode.insertBefore(newNode, referenceNode.nextSibling);
      }
    }, {
      key: "managePlaceholderLists",
      value: function managePlaceholderLists(e) {
        var _this2 = this;

        var actions = this.analysePlaceHolderSituation(e);
        actions.forEach(function (action) {
          switch (action) {
            case 'add':
              _this2.cleanupPlaceholderLists();

              _this2.addPlaceholderList();

              break;

            case 'cleanup':
              _this2.cleanupPlaceholderLists();

              break;
          }
        });
      }
    }, {
      key: "targetedNodeIsPlaceholder",
      value: function targetedNodeIsPlaceholder() {
        return this.targetedNode.nodeName === 'UL' && this.targetedNode.classList.contains(this.classNames.placeholder);
      }
    }, {
      key: "analysePlaceHolderSituation",
      value: function analysePlaceHolderSituation(e) {
        if (!this.targetedNode || this.areNested(this.targetedNode, this.draggedNode)) {
          return [];
        }

        var actions = [];

        if (!this.cursorIsIndentedEnough() || this.mouseIsTooCloseToTop()) {
          if (!this.targetedNodeIsPlaceholder()) {
            actions.push('cleanup');
          }
        } else if (this.targetedNode !== this.draggedNode && this.targetedNode.nodeName === 'LI' && !this.targetedNode.querySelectorAll('ul').length) {
          actions.push('add');
        }

        return actions;
      }
    }, {
      key: "addPlaceholderList",
      value: function addPlaceholderList() {
        var list = this.getPlaceholderList();
        list.style.minHeight = '0';
        this.targetedNode.appendChild(list);
        list.style.transition = 'min-height ease .2s';
        list.style.minHeight = "".concat(this.draggedNode.offsetHeight, "px");
      }
    }, {
      key: "targetNodeIsIdentified",
      value: function targetNodeIsIdentified() {
        return !!this.targetedNode;
      }
    }, {
      key: "canBeDropped",
      value: function canBeDropped() {
        var result = true;
        result &= this.targetNodeIsIdentified() && this.targetedNode !== this.draggedNode;
        result &= this.targetNodeIsIdentified() && !(this.targetedNode.nodeName === 'UL' && this.targetedNode.querySelectorAll('li').length);
        result &= !this.areNested(this.targetedNode, this.draggedNode);
        return result;
      }
    }, {
      key: "cleanupPlaceholderLists",
      value: function cleanupPlaceholderLists() {
        var _this3 = this;

        this.getSortableList().querySelectorAll('ul').forEach(function (ul) {
          if (!ul.querySelectorAll('li').length) {
            ul.remove();
          } else if (ul.classList.contains(_this3.classNames.placeholder)) {
            ul.classList.remove(_this3.classNames.placeholder);
            ul.style.minHeight = 'auto';
            ul.dataset.id = ul.parentNode.dataset.id;
          }
        });
      }
    }, {
      key: "initPlaceholderList",
      value: function initPlaceholderList() {
        this.placeholderUl = document.createElement('ul');
        this.placeholderUl.classList.add(this.classNames.placeholder);
      }
    }, {
      key: "getPlaceholderList",
      value: function getPlaceholderList() {
        this.placeholderInUse = this.placeholderUl.cloneNode(true);
        return this.placeholderInUse;
      }
    }]);

    return nestedSort;
  }();

  return nestedSort;

})));
